@@ -2,7 +2,6 @@
 # -*- coding: utf-8 -*-
 '''
 Author: zhangkai
-Email: zhangkai@cmcm.com
 Last modified: 2018-01-04 20:16:00
 '''
 import asyncio
@@ -18,6 +17,7 @@ import os
 import random
 import re
 import socket
+import string
 import threading
 import time
 import urllib.parse
@@ -38,8 +38,10 @@ from urllib.parse import urlunparse
 import chardet
 import pycurl
 import requests
+import tornado.curl_httpclient
 import yaml
 from bs4 import BeautifulSoup
+from lxml import etree
 from requests.cookies import MockRequest
 from requests.cookies import MockResponse
 from requests.structures import CaseInsensitiveDict
@@ -56,15 +58,14 @@ from tornado.httpclient import HTTPClient
 from tornado.httpclient import HTTPRequest
 from tornado.httpclient import HTTPResponse
 
+from .cached_property import cached_property
 from .utils import Dict
 from .utils import DictWrapper
-from .utils import property_wraps
 from .utils import to_bytes
 from .utils import to_str
 from .utils import tqdm
-logging.getLogger("requests").setLevel(logging.WARNING)
-# requests.packages.urllib3.disable_warnings()
 
+logging.getLogger("requests").setLevel(logging.WARNING)
 
 __all__ = ['Response', 'Request', 'Chrome']
 
@@ -112,12 +113,13 @@ class Response:
     def __init__(self, *args, **kwargs):
         if args and isinstance(args[0], HTTPResponse):
             self.url = args[0].effective_url
-            for key in ['code', 'reason', 'headers', 'body', 'request', 'request_time']:
+            self.time = args[0].request_time
+            for key in ['code', 'reason', 'headers', 'body', 'request']:
                 setattr(self, key, getattr(args[0], key))
         else:
             kwargs.setdefault('body', b'')
-            for key in ['code', 'reason', 'headers', 'body', 'request', 'request_time',
-                        'url', 'cookies', 'encoding']:
+            for key in ['code', 'reason', 'headers', 'body', 'request',
+                        'url', 'time', 'cookies', 'encoding']:
                 if key in kwargs:
                     setattr(self, key, kwargs[key])
 
@@ -127,6 +129,9 @@ class Response:
                 for cookie in self.headers.get_list('Set-Cookie'):
                     sc = SimpleCookie(cookie)
                     self.cookies.update(dict(map(lambda x: (x[0], x[1].value), sc.items())))
+            self.headers = CaseInsensitiveDict(self.headers.items())
+        elif not hasattr(self, 'headers'):
+            self.headers = CaseInsensitiveDict()
 
     @staticmethod
     def _decompress_gzip(body):
@@ -156,8 +161,7 @@ class Response:
         except Exception:
             return body
 
-    @property
-    @property_wraps
+    @cached_property
     def encoding(self):
         encoding = None
         content_type = self.headers.get('content-type')
@@ -167,14 +171,9 @@ class Response:
                 encoding = params['charset'].strip("'\"")
             if encoding is None:
                 encoding = chardet.detect(self.body)['encoding']
-        return encoding
+        return 'gbk' if encoding and encoding.lower() == 'gb2312' else encoding
 
-    @encoding.setter
-    def encoding(self, value):
-        self._encoding = value
-
-    @property
-    @property_wraps
+    @cached_property
     def text(self):
         if self.encoding:
             return self.body.decode(self.encoding)
@@ -187,8 +186,11 @@ class Response:
     def soup(self, features='html5lib', **kwargs):
         return BeautifulSoup(self.body, features=features, **kwargs)
 
+    def html(self, **kwargs):
+        return etree.HTML(self.soup().renderContents())
+
     def __repr__(self):
-        return f'<Response [{self.code} {self.reason}]>'
+        return f'<Response({self.code}) [{self.reason}]>'
 
 
 class RequestMeta(type):
@@ -216,10 +218,10 @@ class RequestMeta(type):
 
 class UserAgent:
 
-    def __init__(self):
-        filename = Path(__file__).absolute().parent / 'user_agent.json'
-        with open(filename) as fp:
-            self._user_agents = json.load(fp)
+    def __init__(self, config='user_agent.json'):
+        filename = Path(config)
+        if filename.exists():
+            self._user_agents = json.load(open(filename))
             self._total_agents = reduce(lambda x, y: x + y, self._user_agents.values())
 
     def __getattr__(self, key):
@@ -236,7 +238,7 @@ class BaseRequest(metaclass=RequestMeta):
         files: dict, { key: filename or file descriptor }
         proxy: str, 'http://user:password@114.112.93.35:8080'
         '''
-        self.proxy = kwargs.get('proxy')
+        self.proxy = kwargs.get('proxy', os.environ.get('http_proxy'))
         self.cookie = kwargs.get('cookie')
         self.cookies = kwargs.get('cookies', {})
         self.timeout = kwargs.get('timeout', 30)
@@ -245,9 +247,6 @@ class BaseRequest(metaclass=RequestMeta):
         self.raise_error = kwargs.get('raise_error', False)
         self.progress = kwargs.get('progress', False)
         self.logger = logging.getLogger()
-        self.ua = UserAgent()
-        # if not self.logger.handlers:
-        #     self.logger.addHandler(logging.NullHandler())
 
         self.headers = {
             'accept': '*/*',
@@ -256,8 +255,7 @@ class BaseRequest(metaclass=RequestMeta):
             'accept-language': 'zh-CN,zh;q=0.8,en;q=0.6',
             'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.75 Safari/537.36'
         }
-        if kwargs.get('headers'):
-            self.headers.update(kwargs['headers'])
+        self.headers.update(kwargs.get('headers', {}))
 
         if self.cookie:
             root = Path(self.cookie).parent
@@ -274,16 +272,13 @@ class BaseRequest(metaclass=RequestMeta):
         self.headers['user-agent'] = 'Mozilla/5.0 (compatible; Baiduspider/2.0; +http://www.baidu.com/search/spider.html)'
 
     def set_chrome_ua(self):
-        self.headers['user-agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.81 Safari/537.36'
+        self.headers['user-agent'] = 'Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.87 Mobile Safari/537.36'
 
     def set_mobile_ua(self):
         self.headers['user-agent'] = 'Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1'
 
-    def set_random_ua(self):
-        self.headers['user-agent'] = self.ua.random
-
-    def _prepare(self, url, **kwargs):
-        url = to_str(url)
+    def _prepare(self, url, kwargs):
+        url = urllib.parse.quote(to_str(url), safe=string.printable)
         kwargs = copy.copy(kwargs)
         headers = copy.copy(self.headers)
         headers.update(kwargs.get('headers', {}))
@@ -339,58 +334,66 @@ class BaseRequest(metaclass=RequestMeta):
             url = 'http:' + url
         if kwargs.pop('autoreferer', False):
             ret = urllib.parse.urlparse(url)
-            headers.update({
-                'referer': f'{ret.scheme}://{ret.netloc}',
-            })
+            headers.update({'referer': f'{ret.scheme}://{ret.netloc}'})
         kwargs['headers'] = headers
         kwargs['cookies'] = cookies
         return url, kwargs
 
+    def _finish(self, kwargs):
+        if hasattr(self, '_pbar'):
+            self._pbar.close()
+        for key, value in kwargs.get('files', []):
+            if not value[1].closed:
+                value[1].close()
+
     def _request(self, url, **kwargs):
         pass
 
-    def _retry(self, url, retry, kwargs, reason=None):
+    def _retry(self, url, kwargs, retry):
+        _kwargs = copy.deepcopy(kwargs)
         try:
-            if retry < 0:
-                resp = Response(url=url, code=599, reason=str(reason))
-            else:
-                resp = self._request(url, **kwargs)
-            for key, value in kwargs.get('files', []):
-                if not value[1].closed:
-                    value[1].close()
-            return resp
+            return self._request(url, **kwargs)
         except Exception as e:
-            self.logger.error(f'retry: {retry} {e}')
-            time.sleep(self.sleep)
-            return self._retry(url, retry - 1, kwargs, e)
+            self.logger.error(f'url: {url}, retry: {retry}, exception: {e}')
+            if retry <= 0:
+                return Response(url=url, code=599, reason=str(e))
+            else:
+                time.sleep(self.sleep)
+                return self._retry(url, _kwargs, retry - 1)
 
     def request(self, url, **kwargs):
-        url, kwargs = self._prepare(url, **kwargs)
-        if self.raise_error:
+        raise_error = kwargs.pop('raise_error', self.raise_error)
+        retry = kwargs.pop('retry', self.retry)
+        if raise_error:
             return self._request(url, **kwargs)
         else:
-            return self._retry(url, self.retry, kwargs)
+            return self._retry(url, kwargs, retry)
 
     def download(self, url, filename=None, shards=1, **kwargs):
-        url, kwargs = self._prepare(url, **kwargs)
+        url, kwargs = self._prepare(url, kwargs)
         kwargs['headers']['Range'] = 'bytes=0-0'
         resp = self._request(url, **kwargs)
-        file_size = int(resp.headers['Content-Range'].split('/')[-1])
-
-        filename = filename or Path(url).name
+        filesize = int(resp.headers['Content-Range'].split('/')[-1])
+        filename = filename or url.split('/')[-1]
+        lock = threading.Lock()
         if shards > 1:
             with open(filename, 'wb') as fp:
-                fp.truncate(file_size)
-            block_size = int(math.ceil(file_size / shards))
-            with open(filename, 'rb+') as fp:
-                lock = threading.Lock()
-                func = partial(self._download_thread, url, fp, lock, file_size, block_size, kwargs)
-                with ThreadPoolExecutor(shards) as executor:
-                    executor.map(func, range(shards))
+                fp.truncate(filesize)
+            fp = open(filename, 'rb+')
+            block = int(math.ceil(filesize / shards))
+            chunks = [(i * block, min(filesize, (i + 1) * block) - 1) for i in range(shards)]
+            args = list(zip(*chunks))
+            func = partial(self._download, url, kwargs, fp, lock)
+            with ThreadPoolExecutor(shards) as executor:
+                executor.map(func, *args)
         else:
-            first_byte = os.path.getsize(filename) if os.path.exists(filename) else 0
-            kwargs['headers']['Range'] = f'bytes={first_byte}-{file_size}'
-            self._download(url, filename, file_size, first_byte, kwargs)
+            start = os.path.getsize(filename) if os.path.exists(filename) else 0
+            end = filesize - 1
+            fp = open(filename, 'ab')
+            if start < end:
+                kwargs['headers']['Range'] = f'bytes={start}-{end}'
+                self._download(url, kwargs, fp, lock, start, end)
+        fp.close()
 
     def get(self, url, **kwargs):
         return self.request(url, method='get', **kwargs)
@@ -475,8 +478,8 @@ class Requests(BaseRequest):
         if self.proxy:
             self.session.proxies = {'http': self.proxy, 'https': self.proxy}
 
-    def _prepare(self, url, **kwargs):
-        url, kwargs = super()._prepare(url, **kwargs)
+    def _prepare(self, url, kwargs):
+        url, kwargs = super()._prepare(url, kwargs)
         kwargs.setdefault('verify', False)
         chunked = kwargs.pop('chunked', False)
         chunk_size = kwargs.pop('chunk_size', 1048576)
@@ -502,15 +505,13 @@ class Requests(BaseRequest):
 
         return url, kwargs
 
-    def _download_thread(self, url, fp, lock, file_size, block_size, kwargs, thread_id):
+    def _download(self, url, kwargs, fp, lock, start, end):
         kwargs = copy.copy(kwargs)
-        start = block_size * thread_id
-        end = min(block_size * (thread_id + 1), file_size) - 1
-        pbar = tqdm(total=end - start + 1, unit='B', unit_scale=True, desc=f'{thread_id}')
         kwargs['headers']['Range'] = f'bytes={start}-{end}'
         kwargs['stream'] = True
         method = kwargs.pop('method', 'GET')
         resp = self.session.request(method, url, **kwargs)
+        pbar = tqdm(total=end - start + 1, unit='B', unit_scale=True, desc=f'{start}')
         for chunk in resp.iter_content(chunk_size=4096):
             if chunk:
                 with lock:
@@ -520,39 +521,26 @@ class Requests(BaseRequest):
                     pbar.update(incr=len(chunk))
         pbar.close()
 
-    def _download(self, url, filename, file_size, first_byte, kwargs):
-        kwargs['stream'] = True
-        pbar = tqdm(total=file_size, initial=first_byte, unit='B', unit_scale=True)
-        method = kwargs.pop('method', 'GET')
-        resp = self.session.request(method, url, **kwargs)
-        with open(filename, 'ab') as fp:
-            for chunk in resp.iter_content(chunk_size=4096):
-                if chunk:
-                    fp.write(chunk)
-                    pbar.update(incr=len(chunk))
-        pbar.close()
-
     def _request(self, url, **kwargs):
+        url, kwargs = self._prepare(url, kwargs)
         request_cookies = requests.utils.dict_from_cookiejar(self.session.cookies)
         request_cookies.update(kwargs['cookies'])
         method = kwargs.pop('method', 'GET')
         resp = self.session.request(method, url, **kwargs)
-        if hasattr(self, '_pbar'):
-            self._pbar.close()
-        if kwargs.get('stream') and method == 'GET':
-            pbar = tqdm(total=int(resp.headers.get('Content-Length', -1)), unit='B', unit_scale=True)
+        if kwargs.get('stream'):
+            self._pbar = tqdm(total=int(resp.headers.get('Content-Length', -1)), unit='B', unit_scale=True)
             content = b''
-            for chunk in resp.iter_content(chunk_size=1024):
+            for chunk in resp.iter_content(chunk_size=4096):
                 if chunk:
                     content += chunk
-                    pbar.update(incr=len(chunk))
-            pbar.close()
+                    self._pbar.update(incr=len(chunk))
         else:
             content = resp.content
+        self._finish(kwargs)
+        self.save_cookie()
 
         kwargs['method'] = method
         response_cookies = requests.utils.dict_from_cookiejar(resp.cookies)
-        self.save_cookie()
         request = Dict(
             url=url,
             headers=kwargs['headers'],
@@ -568,7 +556,7 @@ class Requests(BaseRequest):
             cookies=response_cookies,
             reason=resp.reason,
             request=request,
-            request_time=resp.elapsed.microseconds / 1e6,
+            time=resp.elapsed.microseconds / 1e6,
         )
         return response
 
@@ -627,7 +615,8 @@ class Pycurl(BaseRequest):
         curl.setopt(pycurl.VERBOSE, 0)
         curl.setopt(pycurl.SSL_VERIFYPEER, 0)
         curl.setopt(pycurl.SSL_VERIFYHOST, 0)
-        curl.setopt(pycurl.AUTOREFERER, 1)
+        # curl.setopt(pycurl.AUTOREFERER, 1)
+        # 导致未知的403错误 https://video.wanmeikk.me/hls/a02689e5-d0f6-4a55-a590-75119b142a21/d6MJXL_kK9bOc1cTqAs3vVOWDA6wepIKOln9hG_oIT2vMXt81bvAsSaNmOVnaXAcM2BsEd87CSN9VlFIrLZuSw==.ts
         curl.setopt(pycurl.URL, native_str(url))
 
         curl.setopt(pycurl.TRANSFER_ENCODING, int(kwargs.get('chunked', False)))
@@ -734,13 +723,11 @@ class Pycurl(BaseRequest):
                 headers.parse_line(header_line)
         return reason, headers
 
-    def _download_thread(self, url, fp, lock, file_size, block_size, kwargs, thread_id):
-        kwargs = copy.copy(kwargs)
-        start = block_size * thread_id
-        end = min(block_size * (thread_id + 1), file_size) - 1
-        kwargs['headers']['Range'] = f'bytes={start}-{end}'
-        pbar = tqdm(total=end - start + 1, unit='B', unit_scale=True, desc=f'{thread_id}')
+    def _download(self, url, kwargs, fp, lock, start, end):
         with lock:
+            kwargs = copy.copy(kwargs)
+            kwargs['headers']['Range'] = f'bytes={start}-{end}'
+            pbar = tqdm(total=end - start + 1, unit='B', unit_scale=True, desc=f'{start}')
             fp.seek(start)
             self._curl_setup(self.c, url, **kwargs)
             self.c.setopt(pycurl.WRITEFUNCTION, fp.write)
@@ -749,24 +736,13 @@ class Pycurl(BaseRequest):
             self.c.perform()
             pbar.close()
 
-    def _download(self, url, filename, file_size, first_byte, kwargs):
-        self._curl_setup(self.c, url, **kwargs)
-        fp = open(filename, 'ab')
-        self.c.setopt(pycurl.WRITEFUNCTION, fp.write)
-        self.c.setopt(pycurl.NOPROGRESS, 0)
-        pbar = tqdm(total=file_size, initial=first_byte, unit='B', unit_scale=True)
-        self.c.setopt(pycurl.PROGRESSFUNCTION, lambda a, b, c, d: pbar.update(first_byte + int(b), int(a)))
-        self.c.perform()
-        fp.close()
-        pbar.close()
-
     def _request(self, url, **kwargs):
+        url, kwargs = self._prepare(url, kwargs)
         self._curl_setup(self.c, url, **kwargs)
         request_cookies = requests.utils.dict_from_cookiejar(self.cookiejar)
         request_cookies.update(kwargs['cookies'])
         self.c.perform()
-        if hasattr(self, '_pbar'):
-            self._pbar.close()
+        self._finish(kwargs)
 
         reason, headers = self._parse_headers(self.h.getvalue())
         request_time = self.c.getinfo(pycurl.TOTAL_TIME)
@@ -792,53 +768,53 @@ class Pycurl(BaseRequest):
             headers=headers,
             body=Response._decompress(headers, body),
             request=request,
-            request_time=request_time,
+            time=request_time,
         )
         return resp
 
 
 class TornadoClient(Pycurl):
 
-    def __init__(self, max_clients=50, **kwargs):
+    def __init__(self, max_clients=5, **kwargs):
         ''' This cannot handle 301/302 redirection cookies
         '''
         super().__init__(**kwargs)
         self.http = CurlAsyncHTTPClient(max_clients=max_clients)
 
-    async def _retry(self, url, retry, kwargs):
+    async def _retry(self, url, kwargs, retry):
+        _kwargs = copy.deepcopy(kwargs)
         try:
             return await self._request(url, **kwargs)
-        except Exception as e:
-            self.logger.error(f'retry: {retry} {e}')
+        except tornado.curl_httpclient.CurlError as e:
+            self.logger.error(f'url: {url}, retry: {retry}, exception: {e}')
             if retry <= 0:
-                return Response(url=url, code=e.code, reason=str(e))
-            await asyncio.sleep(self.sleep)
-            return await self._retry(url, retry - 1, kwargs)
+                return Response(url=url, code=e.code, reason=e.message)
+            else:
+                await asyncio.sleep(self.sleep)
+                return await self._retry(url, _kwargs, retry - 1)
 
     async def _request(self, url, **kwargs):
-        url, kwargs = self._prepare(url, **kwargs)
+        url, kwargs = self._prepare(url, kwargs)
         curl_callback = partial(self._curl_setup, url=url, **kwargs)
-
         resp = Response(await self.http.fetch(url,
-                                              prepare_curl_callback=curl_callback,
-                                              raise_error=self.raise_error))
-        if hasattr(self, '_pbar'):
-            self._pbar.close()
+                                              raise_error=False,
+                                              prepare_curl_callback=curl_callback))
+        self._finish(kwargs)
         self.save_cookie(resp)
         return resp
 
     async def download(self, url, filename=None, shards=1, **kwargs):
         filename = filename or Path(url).name
         if shards > 1:
-            url, kwargs = self._prepare(url, **kwargs)
+            url, kwargs = self._prepare(url, kwargs)
             kwargs['headers']['Range'] = 'bytes=0-0'
             resp = await self._request(url, **kwargs)
-            file_size = int(resp.headers['Content-Range'].split('/')[-1])
-            block_size = int(math.ceil(file_size / shards))
+            filesize = int(resp.headers['Content-Range'].split('/')[-1])
+            block = int(math.ceil(filesize / shards))
             futures = []
             for i in range(shards):
-                start = block_size * i
-                end = min(block_size * (i + 1), file_size) - 1
+                start = block * i
+                end = min(block * (i + 1), filesize) - 1
                 _kwargs = copy.deepcopy(kwargs)
                 _kwargs['headers']['Range'] = f'bytes={start}-{end}'
                 futures.append(self._request(url, **_kwargs))
@@ -881,7 +857,7 @@ class Fetcher(BaseRequest):
         return fetch
 
     def _request(self, url, **kwargs):
-        url, kwargs = self._prepare(url, **kwargs)
+        url, kwargs = self._prepare(url, kwargs)
         body = self.parse_option(url, **kwargs)
         config = {
             'follow_redirects': False,
@@ -903,7 +879,6 @@ class Fetcher(BaseRequest):
             body=doc.content,
             code=doc.status_code,
             cookies=doc.cookies,
-            encoding=Response._get_encoding(doc.headers, doc.body),
             reason=doc.error or 'OK',
         )
         return response
@@ -927,8 +902,10 @@ class Chrome(webdriver.Chrome):
             options.add_argument('headless')
         if proxy:
             options.add_argument(f'proxy-server={proxy}')
-        if mobile:
+        if mobile == 'ios':
             options.add_argument('user-agent="Mozilla/5.0 (iPhone; CPU iPhone OS 9_1 like Mac OS X) AppleWebKit/601.1.46 (KHTML, like Gecko) Version/9.0 Mobile/13B143 Safari/601.1"')
+        elif mobile == 'android':
+            options.add_argument('user-agent="Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.87 Mobile Safari/537.36"')
         d = DesiredCapabilities.CHROME
         d['loggingPrefs'] = {'performance': 'ALL'}
         super().__init__(chrome_options=options, desired_capabilities=d)
@@ -988,12 +965,3 @@ def Request(lib='requests', **kwargs):
         'tornado': TornadoClient,
     }
     return module[lib](**kwargs)
-
-
-if __name__ == '__main__':
-    loop = asyncio.get_event_loop()
-    http = Request(lib='tornado')
-
-    async def fetch():
-        await http.get('https://www.baidu.com')
-    loop.run_until_complete(fetch())
