@@ -24,7 +24,7 @@ def Logger(filename=None, name=None, level='INFO', stream=True):
     logger.propagate = False
     logger.handlers = []
     datefmt = '%Y-%m-%d %H:%M:%S'
-    fmt = '%(color)s[%(levelname)s %(asctime)s %(module)s:%(lineno)d]%(end_color)s %(message)s'
+    fmt = '%(color)s[%(levelname)s %(asctime)s pid:%(process)d %(module)s:%(lineno)d]%(end_color)s %(message)s'
     if stream:
         hdlr = logging.StreamHandler()
         hdlr.setFormatter(LogFormatter(fmt=fmt, datefmt=datefmt))
