@@ -9,34 +9,61 @@ import asyncio
 import collections
 import datetime
 import decimal
-import functools
+import inspect
 import json
 import math
-import os
-import smtplib
+import queue
 import socket
 import threading
 import time
-from email.mime.application import MIMEApplication
-from email.mime.multipart import MIMEMultipart
-from email.mime.text import MIMEText
-from email.utils import COMMASPACE
-from email.utils import formatdate
+from pathlib import Path
 from socket import inet_aton
 from socket import inet_ntoa
 from struct import pack
 from struct import unpack
 
-import aiosmtplib
-import numpy as np
-import pymongo
+import requests
+import yaml
+from orderedset import OrderedSet
+from pymongo.cursor import Cursor
 from tqdm import tqdm
 
 
+class Queue(queue.Queue):
+
+    def _init(self, maxsize):
+        self.queue = OrderedSet()
+
+    def _put(self, item):
+        self.queue.add(item)
+
+    def _get(self):
+        return self.queue.pop(last=False)
+
+    def __contains__(self, item):
+        with self.mutex:
+            return item in self.queue
+
+
+class AioQueue(asyncio.Queue):
+
+    def _init(self, maxsize):
+        self._queue = OrderedSet()
+
+    def _put(self, item):
+        self._queue.add(item)
+
+    def _get(self):
+        return self._queue.pop(last=False)
+
+    def __contains__(self, item):
+        return item in self.queue
+
+
 class tqdm(tqdm):
 
     def __init__(self, *args, **kwargs):
-        if args and isinstance(args[0], pymongo.cursor.Cursor):
+        if args and isinstance(args[0], Cursor):
             kwargs.setdefault('total', args[0].count())
         super().__init__(*args, **kwargs)
 
@@ -49,6 +76,10 @@ class tqdm(tqdm):
             super().update(incr)
 
 
+async def awaitable(ret):
+    return await ret if inspect.isawaitable(ret) else ret
+
+
 def floor(number, ndigits=0):
     '''当ndigits大于等于number的小数点位数时，直接返回
     '''
@@ -81,15 +112,22 @@ def to_bytes(*args):
     return result[0] if len(args) == 1 else result
 
 
-def get_ip():
-    try:
-        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-        s.connect(('8.8.8.8', 80))
-        return s.getsockname()[0]
-    except Exception:
-        return '127.0.0.1'
-    finally:
-        s.close()
+def get_ip(local=True):
+    if local:
+        try:
+            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
+            s.connect(('8.8.8.8', 80))
+            return s.getsockname()[0]
+        except Exception:
+            return '127.0.0.1'
+        finally:
+            s.close()
+    else:
+        try:
+            resp = requests.get('https://api.ishield.cn/ip', timeout=3)
+            return resp.text
+        except Exception:
+            return '127.0.0.1'
 
 
 def connect(ip, port):
@@ -119,26 +157,6 @@ def int2str(int_time):
     return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(int_time))
 
 
-def property_wraps(method):
-    @functools.wraps(method)
-    def wrapper(self, *args, **kwargs):
-        key = f'_{method.__name__}'
-        if not hasattr(self, key):
-            setattr(self, key, method(self, *args, **kwargs))
-        return getattr(self, key)
-    return wrapper
-
-
-def time_wraps(method):
-    @functools.wraps(method)
-    def wrapper(*args, **kwargs):
-        start_time = time.time()
-        result = method(*args, **kwargs)
-        print(f'{method.__name__} cost time: {time.time() - start_time:.5f}')
-        return result
-    return wrapper
-
-
 class JSONEncoder(json.encoder.JSONEncoder):
     '''针对某些不能序列化的类型如datetime，使用json.dumps(data, cls=JSONEncoder)
     '''
@@ -146,16 +164,16 @@ class JSONEncoder(json.encoder.JSONEncoder):
     def default(self, obj):
         if isinstance(obj, set):
             return list(obj)
-        elif isinstance(obj, np.integer):
-            return int(obj)
-        elif isinstance(obj, (np.floating, decimal.Decimal)):
+        elif isinstance(obj, decimal.Decimal):
             return float(obj)
-        elif isinstance(obj, np.ndarray):
-            return obj.tolist()
         elif isinstance(obj, datetime.datetime):
             return obj.strftime('%Y-%m-%d %H:%M:%S')
         elif isinstance(obj, datetime.date):
             return obj.strftime('%Y-%m-%d')
+        elif isinstance(obj, Path):
+            return str(obj)
+        elif hasattr(obj, 'tolist') and callable(obj.tolist):
+            return obj.tolist()
         try:
             return super().default(obj)
         except Exception:
@@ -253,94 +271,11 @@ def DictUnwrapper(doc):
     return doc
 
 
-class EmailBase:
-
-    def __init__(self, sender=None, smtp=None, user=None, pwd=None):
-        self.sender = sender or os.environ.get('EMAIL_SENDER')
-        self.smtp = smtp or os.environ.get('EMAIL_SMTP')
-        self.user = user or os.environ.get('EMAIL_USER')
-        self.pwd = pwd or os.environ.get('EMAIL_PWD')
-
-    def pack(self, receivers, title=None, content=None, files=None, cc=None):
-        msg = MIMEMultipart()
-
-        if content:
-            mime = MIMEText(content, 'html', 'utf-8')
-            msg.attach(mime)
-
-        if files:
-            if isinstance(files, (str, bytes)):
-                files = [files]
-            for i, fname in enumerate(files):
-                '''
-                _type, _ = mimetypes.guess_type(fname)
-                if _type:
-                    _type = _type.split('/')
-                    att = MIMEBase(_type[0], _type[1], filename=fname)
-                else:
-                    att = MIMEBase('application', 'octet-stream')
-                with open(fname, 'rb') as fp:
-                    att.set_payload(fp.read())
-                encoders.encode_base64(att)
-                att['Content-ID'] = str(i)
-                att["Content-Disposition"] = f'attachment; filename="{ os.path.basename(fname) }"'
-                '''
-                att = MIMEApplication(open(fname, 'rb').read())
-                att.add_header('Content-ID', str(i))
-                # att.add_header('X-Attachment-Id', str(i))
-                att.add_header('Content-Type', 'application/octet-stream')
-                att.add_header('Content-Disposition', 'attachment', filename=os.path.basename(fname))
-                msg.attach(att)
-
-        if cc:
-            if not isinstance(cc, list):
-                cc = [cc]
-            msg['cc'] = COMMASPACE.join(cc)
-
-        msg['subject'] = title
-        msg['date'] = formatdate(localtime=True)
-        msg['from'] = self.sender
-        if not isinstance(receivers, list):
-            receivers = [receivers]
-        msg['to'] = COMMASPACE.join(receivers)
-        return msg
+def yaml_load(stream, **kwargs):
+    doc = yaml.load(stream, **kwargs)
+    return DictWrapper(doc)
 
 
-class Email(EmailBase):
-
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        self.client = smtplib.SMTP()
-        # self.client = smtplib.SMTP('localhost')
-        # self.sender = self.client.local_hostname
-
-    def send(self, *args, **kwargs):
-        msg = self.pack(*args, **kwargs)
-        self.client.connect(self.smtp)
-        self.client.docmd('ehlo', self.smtp)
-        self.client.login(self.user, self.pwd)
-        self.client.send_message(msg)
-        self.client.quit()
-
-
-class AioEmail(EmailBase):
-
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        loop = asyncio.get_event_loop()
-        self.client = aiosmtplib.SMTP(loop=loop)
-        # self.client = smtplib.SMTP('localhost')
-        # self.sender = self.client.hostname
-
-    async def send(self, *args, **kwargs):
-        msg = self.pack(*args, **kwargs)
-        await self.client.connect(self.smtp)
-        await self.client.login(self.user, self.pwd)
-        await self.client.send_message(msg)
-        await self.client.quit()
-
-
-if __name__ == '__main__':
-    email = AioEmail()
-    loop = asyncio.get_event_loop()
-    loop.run_until_complete(email.send('mpbase1@163.com', 'hello', 'world'))
+def yaml_dump(doc, stream=None, default_flow_style=False, allow_unicode=True, **kwargs):
+    doc = DictUnwrapper(doc)
+    return yaml.dump(doc, stream=stream, default_flow_style=default_flow_style, allow_unicode=allow_unicode, **kwargs)
